--[[Simplistic utility for compressing and decompressing tables of strings into strings
it strips whatever formalities and uses non-printing characters to separate key-value
so that at first table is like (let @ signify non-printing separator):
    key1@value1@key2@value2key3@value3

--then it splits all keys and values into tables of characters:
    --{k,e,y,1},{v,a,l,u,e,1}
and tries to find repeating sequences of at least two characters,
to each sequence a price is assigned, defined by (frequency - 1)*(length-1) - 2
sequence with highest price is replaced by non-print char
and subsequencies are recalculated 

in the head it creates a dictionary which is pairs of string - non-print code
that dictionary is ended by another non-print
concatenation of the dictionary and compressed string is then returned
]]
local insert=table.insert
local string=string
local word
word={
    new=function(str)
        return setmetatable({
            text=assert(tostring(str)),
            },
            word)
    end,
    __index={
        gsub=function(me,pattern,replacement)
            me.text=me.text:gsub(pattern,replacement)
            return me.text
        end,
        sub=function(me,i,j)
            return me.text:sub(i,j)
        end,
        subsequencies=function(me)
            local text=me.text
            local n=#text
            local i,j=1,1
            return function()
                j=j+1
                if j>n then
                    i=i+1
                    j=i+1
                end
                local res=me:sub(i,j)
                return not (res=='') and res or nil
            end
        end
    },
    __tostring=function(me) return me.text end,
    __concat=function(me,other) return tostring(me)..tostring(other) end
}

local tab2words=function(tab)
    local words={}
    for k,v in pairs(tab) do
        insert(words,word.new(k))
        insert(words,word.new(v))
    end
    return words
end
local seq_frequencies=function(words)
    local freqs={}
    for _,word in pairs(words) do
        for sub in word:subsequencies() do
            freqs[sub]=(freqs[sub] or 0) +1
        end
    end
    return freqs
end

local rawstring=function(str)
    return str:gsub([=[([%^%$%(%)%%%.%[%]%*%+%-%?])]=],"%%%1")
end
local find_valuable=function(frequencies)
    local best_price=0
    local str=nil
    for k,fr in pairs(frequencies) do
        local price= (#k-1)*(fr-1)-2
        if price>best_price then
            best_price=price
            str=k
        end
    end
    if not str then return end
    return rawstring(str)
    --return str
end
local replace_valuable=function(words,pattern,char)
    for _,word in pairs(words) do
        --print(word,pattern)
        word:gsub(pattern,char)
        --print(word)
    end
end

local sep=string.char(127)
local dic_end=string.char(126)
local start_symbol=string.char(129)

local next_char=function(char)
    return char.char(char:byte()+1)
end
local concat_words=function(words,separator)
    local t={}
    for i=1,#words do 
        insert(t,words[i].text)
    end
    return table.concat(t,separator)
end
local reverse=function(arr)
    local res={}
    for i=#arr,1,-1 do
        insert(res,arr[i])
    end
    return res
end

local c=3
local tic= function() c=c-1 assert(c>0) end
local compressor={
    compress=function(tab)
        local words=tab2words(tab)
        local valuable=find_valuable(seq_frequencies(words))
        local dict={}
        local symbol=start_symbol
        while (valuable and (symbol:byte()<255)) do
            --print('\n\nloop',valuable)
            insert(dict,valuable)
            insert(dict,symbol)
            --print(dict)
            --print(words)
            replace_valuable(words,valuable,symbol)
            --print(words)
            valuable=find_valuable(seq_frequencies(words))
            --tic()
            symbol=next_char(symbol)
        end
        local rev_dict=reverse(dict)
        local cc=table.concat
        return cc(rev_dict,sep)..dic_end..concat_words(words,sep)
    end,
    uncompress=function(str)
        --str=rawstring(str)
        local dict,tab=str:match("(.*)"..dic_end.."(.*)")
        --print('dict\n',dict,'\ntab\n',tab)
        local pair="([^"..sep.."]+)"..sep.."([^"..sep.."]+)"
        local _,right,sym,val=dict:find(pair)
        --print('sym\n',sym,'\nval\n',val)
        while val do
            local v=val:gsub("%%","")
            dict=dict:sub(right+1):gsub(sym,v)
            --print(val)
            --print(val:gsub("%","%%%%"))
            --print(rawstring(val))
            --print(sym)
            --print(val)
            
            --print(v)
            --tab=tab:gsub(sym,val)
            tab=tab:gsub(sym,v)
            --print(tab)
            _,right,sym,val=dict:find(pair)
            --print('sym\n',sym,'\nval\n',val)
        end
        local res={}
        for k,v in tab:gmatch(pair) do
            res[k]=v
        end
        return res
    end,
}

local function test(test_table)
    local t=test_table or {
      ["cargo-drone-Adil"] = "cargo-drone-Adil-capsule",
      ["construction-drone-Adil"] = "construction-drone-Adil-capsule",
      defender = "defender-capsule",
      destroyer = "destroyer-capsule",
      distractor = "distractor-capsule",
      ["exterminator-Adil"] = "exterminator-Adil-capsule",
      ["protector-Adil"] = "protector-Adil-capsule"
    }
    local a=compressor.compress(t)
    --print(a)
    local q=compressor.uncompress(a)
    for k,v in pairs(q) do
        assert(v==t[k],"Error in uncompressed field "..k)
    end
    for k,v in pairs(t) do
        assert(v==q[k],"Error in compressing field"..k)
    end
    print('test passed')
    print('compression '..#a..' against '..#((require'serpent').dump(t)))
end

return compressor,test
